==> sane-0.1.tar.gz: sane-0.1/backend/metadl.c <==
/* sane - Scanner Access Now Easy.
   Copyright (C) 1996 David Mosberger-Tang
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   This file implements a dynamic linking based SANE meta backend.  It
   allows managing an arbitrary number of SANE backends by using
   dynamic linking to load backends on demand.  */

#include <alloca.h>
#include <dlfcn.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>

#include <sane.h>
#include <sanei.h>

#define PATH_METADL_CONFIG	PATH_SANE_CONFIG_DIR "/metadl.conf"

enum SANE_Ops
  {
    OP_EXIT = 0,
    OP_GET_DEVS,
    OP_OPEN,
    OP_CLOSE,
    OP_GET_OPTION_DESC,
    OP_CTL_OPTION,
    OP_GET_PARAMS,
    OP_START,
    OP_READ,
    OP_CANCEL,
    OP_SET_IO_MODE,
    OP_GET_SELECT_FD,
    NUM_OPS
  };

struct backend
  {
    struct backend *next;
    const char *name;
    int loaded;
    void *handle;		/* handle returned by dlopen() */
    void *(*op[NUM_OPS]) ();
  };

struct meta_scanner
  {
    struct backend *be;
    SANE_Handle handle;
  };

static struct backend *first_backend;
static const char *op_name[] =
{
  "sane_exit", "sane_get_devices", "sane_open", "sane_close",
  "sane_get_option_descriptor", "sane_control_option", "sane_get_parameters",
  "sane_start", "sane_read", "sane_cancel",
  "sane_set_io_mode", "sane_get_select_fd"
};

static void *
op_unsupported (void)
{
  return (void *) (long) SANE_STATUS_UNSUPPORTED;
}


static SANE_Status
add_backend (const char *name, struct backend **bep)
{
  struct backend *be = malloc (sizeof (*be));
  if (!be)
    return SANE_STATUS_NO_MEM;

  memset (be, 0, sizeof (*be));
  be->name = strdup (name);
  if (!be->name)
    return SANE_STATUS_NO_MEM;
  be->next = first_backend;
  first_backend = be;
  if (bep)
    *bep = be;
  return SANE_STATUS_GOOD;
}

static SANE_Status
load (struct backend *be)
{
  int mode = getenv ("LD_BIND_NOW") ? RTLD_NOW : RTLD_LAZY;
  SANE_Status (*init) (SANE_Int * version_code);
  SANE_Status status;
  SANE_Int version;
  char *libname;
  size_t len;
  int i;
# define PREFIX "libsane-"
# define POSTFIX ".so"

  /* initialize all ops to "unsupported" so we can "use" the backend
     even if the stuff later in this function fails */
  be->loaded = 1;
  be->handle = 0;
  for (i = 0; i < NUM_OPS; ++i)
    be->op[i] = op_unsupported;

  len = (sizeof (PREFIX) - 1) + strlen (be->name) + (sizeof (POSTFIX) - 1);
  libname = alloca (len);
  if (!libname)
    return SANE_STATUS_NO_MEM;

  strcpy (libname, PREFIX);
  strcat (libname, be->name);
  strcat (libname, POSTFIX);

  be->handle = dlopen (libname, mode);
  if (!be->handle)
    return SANE_STATUS_INVAL;

  init = dlsym (be->handle, "sane_init");
  status = SANE_STATUS_GOOD;
  if (!init || (status = (long) (*init) (&version)) != SANE_STATUS_GOOD
      || SANE_VERSION_MAJOR (version) != V_MAJOR)
    {
      dlclose (be->handle);
      be->handle = 0;
      if (!init || status == SANE_STATUS_GOOD)
	/* no init function or library got bad version code */
	status = SANE_STATUS_INVAL;
      return status;
    }

  /* all is dandy---lookup and fill in backend ops: */
  for (i = 0; i < NUM_OPS; ++i)
    {
      void *(*op) () = dlsym (be->handle, op_name[i]);
      if (op)
	be->op[i] = op;
    }

  return SANE_STATUS_GOOD;

# undef PREFIX
# undef POSTFIX
}

SANE_Status
sane_init (SANE_Int * version_code)
{
  char backend_name[PATH_MAX];
  size_t len;
  FILE *fp;

  if (version_code)
    *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, 0);

  fp = fopen (PATH_METADL_CONFIG, "r");
  if (!fp)
    return SANE_STATUS_GOOD;	/* don't insist on config file */

  while (fgets (backend_name, sizeof (backend_name), fp))
    {
      if (backend_name[0] == '#')	/* ignore line comments */
	continue;
      len = strlen (backend_name);
      if (backend_name[len - 1] == '\n')
	backend_name[--len] = '\0';

      if (!len)
	continue;		/* ignore empty lines */

      add_backend (backend_name, 0);
    }
  fclose (fp);

  return SANE_STATUS_GOOD;
}

SANE_Status
sane_exit (void)
{
  struct backend *be, *next;

  for (be = first_backend; be->next; be = next)
    {
      next = be->next;
      if (be->loaded)
	{
	  (*be->op[OP_EXIT]) ();
	  dlclose (be->handle);
	}
      if (be->name)
	free ((void *) be->name);
      free (be);
    }
  first_backend = 0;
  return SANE_STATUS_GOOD;
}

/* Note that a call to get_devices() implies that we'll have to load
   all backends.  To avoid this, you can call sane_open() directly
   (assuming you know the name of the backend/device).  This is
   appropriate for the command-line interface of SANE, for example.
 */
SANE_Status
sane_get_devices (const SANE_Device *** device_list)
{
  static int devlist_size = 0, devlist_len = 0;
  static const SANE_Device **devlist;
  const SANE_Device **be_list;
  struct backend *be;
  SANE_Status status;
  char *full_name;
  int i, num_devs;
  size_t len;
#define ASSERT_SPACE(n)						\
  {									\
    if (devlist_len + (n) > devlist_size)				\
      {									\
	devlist_size += (n) + 15;					\
	devlist = realloc(devlist, devlist_size * sizeof(devlist[0]));	\
	if (!devlist)							\
	  return SANE_STATUS_NO_MEM;					\
      }									\
  }

  if (devlist)
    for (i = 0; i < devlist_len; ++i)
      free ((void *) devlist[i]);
  devlist_len = 0;

  for (be = first_backend; be; be = be->next)
    {
      if (!be->loaded)
	load (be);

      status = (long) (*be->op[OP_GET_DEVS]) (&be_list);
      if (status != SANE_STATUS_GOOD)
	continue;

      /* count the number of devices for this backend: */
      for (num_devs = 0; be_list[num_devs]; ++num_devs);

      ASSERT_SPACE (num_devs);

      for (i = 0; i < num_devs; ++i)
	{
	  SANE_Device *dev;
	  char *mem;

	  /* create a new device entry with a device name that is the
	     sum of the backend name a colon and the backend's device
	     name: */
	  len = strlen (be->name) + 1 + strlen (be_list[i]->name);
	  mem = malloc (sizeof (*dev) + len + 1);
	  if (!mem)
	    return SANE_STATUS_NO_MEM;

	  full_name = mem + sizeof (*dev);
	  strcpy (full_name, be->name);
	  strcat (full_name, ":");
	  strcat (full_name, be_list[i]->name);

	  dev = (SANE_Device *) mem;
	  dev->name = full_name;
	  dev->vendor = be_list[i]->vendor;
	  dev->model = be_list[i]->model;
	  dev->type = be_list[i]->type;

	  devlist[devlist_len++] = dev;
	}
    }

  /* terminate device list with NULL entry: */
  ASSERT_SPACE (1);
  devlist[devlist_len++] = 0;

  *device_list = devlist;
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_open (SANE_String_Const full_name, SANE_Handle * meta_handle)
{
  const char *be_name, *dev_name;
  struct meta_scanner *s;
  SANE_Handle *handle;
  struct backend *be;
  SANE_Status status;

  dev_name = strchr (full_name, ':');
  if (dev_name)
    {
#ifdef strndupa
      be_name = strndupa (full_name, dev_name - full_name);
#else
      be_name = alloca (dev_name - full_name + 1);
      memcpy (be_name, full_name, dev_name - full_name);
      be_name[dev_name - full_name] = '\0';
#endif
      ++dev_name;		/* skip colon */
    }
  else
    {
      /* if no colon interpret full_name as the backend name; an empty
         backend device name will cause us to open the first device of
         that backend.  */
      be_name = full_name;
      dev_name = "";
    }

  for (be = first_backend; be; be = be->next)
    if (strcmp (be->name, be_name) == 0)
      break;

  if (!be)
    {
      status = add_backend (be_name, &be);
      if (status != SANE_STATUS_GOOD)
	return status;
    }

  if (!be->loaded)
    {
      status = load (be);
      if (status != SANE_STATUS_GOOD)
	return status;
    }

  status = (long) (*be->op[OP_OPEN]) (dev_name, &handle);
  if (status != SANE_STATUS_GOOD)
    return status;

  s = malloc (sizeof (*s));
  if (!s)
    return SANE_STATUS_NO_MEM;

  memset (s, 0, sizeof (*s));
  s->be = be;
  s->handle = handle;
  *meta_handle = s;

  return SANE_STATUS_GOOD;
}

SANE_Status
sane_close (SANE_Handle handle)
{
  struct meta_scanner *s = handle;
  SANE_Status status;

  status = (long) (*s->be->op[OP_CLOSE]) (s->handle);
  free (s);
  return status;
}

const SANE_Option_Descriptor *
sane_get_option_descriptor (SANE_Handle handle, SANE_Int option)
{
  struct meta_scanner *s = handle;
  return (*s->be->op[OP_GET_OPTION_DESC]) (s->handle, option);
}

SANE_Status
sane_control_option (SANE_Handle handle, SANE_Int option,
		     SANE_Action action, void *value, SANE_Word * info)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_CTL_OPTION]) (s->handle, option, action,
					     value, info);
}

SANE_Status
sane_get_parameters (SANE_Handle handle, SANE_Parameters * params)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_GET_PARAMS]) (s->handle, params);
}

SANE_Status
sane_start (SANE_Handle handle)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_START]) (s->handle);
}

SANE_Status
sane_read (SANE_Handle handle, SANE_Byte * data, SANE_Int max_length,
	   SANE_Int * length)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_READ]) (s->handle, data, max_length, length);
}

SANE_Status
sane_cancel (SANE_Handle handle)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_CANCEL]) (s->handle);
}

SANE_Status
sane_set_io_mode (SANE_Handle handle, SANE_Bool non_blocking)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_SET_IO_MODE]) (s->handle, non_blocking);
}

SANE_Status
sane_get_select_fd (SANE_Handle handle, SANE_Int * fd)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_GET_SELECT_FD]) (s->handle, fd);
}

==> sane-0.32.tar.gz: sane-0.32/backend/metadl.c <==
/* sane - Scanner Access Now Easy.
   Copyright (C) 1996 David Mosberger-Tang
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   This file implements a dynamic linking based SANE meta backend.  It
   allows managing an arbitrary number of SANE backends by using
   dynamic linking to load backends on demand.  */

#include <alloca.h>
#include <dlfcn.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>

#include <sane.h>
#include <sanei.h>

#define PATH_METADL_CONFIG	PATH_SANE_CONFIG_DIR "/metadl.conf"

enum SANE_Ops
  {
    OP_EXIT = 0,
    OP_GET_DEVS,
    OP_OPEN,
    OP_CLOSE,
    OP_GET_OPTION_DESC,
    OP_CTL_OPTION,
    OP_GET_PARAMS,
    OP_START,
    OP_READ,
    OP_CANCEL,
    OP_SET_IO_MODE,
    OP_GET_SELECT_FD,
    NUM_OPS
  };

struct backend
  {
    struct backend *next;
    const char *name;
    int loaded;
    void *handle;		/* handle returned by dlopen() */
    void *(*op[NUM_OPS]) ();
  };

struct meta_scanner
  {
    struct backend *be;
    SANE_Handle handle;
  };

static struct backend *first_backend;
static const char *op_name[] =
{
  "sane_exit", "sane_get_devices", "sane_open", "sane_close",
  "sane_get_option_descriptor", "sane_control_option", "sane_get_parameters",
  "sane_start", "sane_read", "sane_cancel",
  "sane_set_io_mode", "sane_get_select_fd"
};

static void *
op_unsupported (void)
{
  return (void *) (long) SANE_STATUS_UNSUPPORTED;
}


static SANE_Status
add_backend (const char *name, struct backend **bep)
{
  struct backend *be = malloc (sizeof (*be));
  if (!be)
    return SANE_STATUS_NO_MEM;

  memset (be, 0, sizeof (*be));
  be->name = strdup (name);
  if (!be->name)
    return SANE_STATUS_NO_MEM;
  be->next = first_backend;
  first_backend = be;
  if (bep)
    *bep = be;
  return SANE_STATUS_GOOD;
}

static SANE_Status
load (struct backend *be)
{
  int mode = getenv ("LD_BIND_NOW") ? RTLD_NOW : RTLD_LAZY;
  SANE_Status (*init) (SANE_Int * version_code);
  SANE_Status status;
  SANE_Int version;
  char *libname;
  size_t len;
  int i;
# define PREFIX "libsane-"
# define POSTFIX ".so"

  /* initialize all ops to "unsupported" so we can "use" the backend
     even if the stuff later in this function fails */
  be->loaded = 1;
  be->handle = 0;
  for (i = 0; i < NUM_OPS; ++i)
    be->op[i] = op_unsupported;

  len = (sizeof (PREFIX) - 1) + strlen (be->name) + (sizeof (POSTFIX) - 1);
  libname = alloca (len);
  if (!libname)
    return SANE_STATUS_NO_MEM;

  strcpy (libname, PREFIX);
  strcat (libname, be->name);
  strcat (libname, POSTFIX);

  be->handle = dlopen (libname, mode);
  if (!be->handle)
    return SANE_STATUS_INVAL;

  init = (SANE_Status (*)(SANE_Int *)) dlsym (be->handle, "sane_init");
  status = SANE_STATUS_GOOD;
  if (!init || (status = (long) (*init) (&version)) != SANE_STATUS_GOOD
      || SANE_VERSION_MAJOR (version) != V_MAJOR)
    {
      dlclose (be->handle);
      be->handle = 0;
      if (!init || status == SANE_STATUS_GOOD)
	/* no init function or library got bad version code */
	status = SANE_STATUS_INVAL;
      return status;
    }

  /* all is dandy---lookup and fill in backend ops: */
  for (i = 0; i < NUM_OPS; ++i)
    {
      void *(*op) () = (void *(*)())dlsym (be->handle, op_name[i]);
      if (op)
	be->op[i] = op;
    }

  return SANE_STATUS_GOOD;

# undef PREFIX
# undef POSTFIX
}

SANE_Status
sane_init (SANE_Int * version_code)
{
  char backend_name[PATH_MAX];
  size_t len;
  FILE *fp;

  if (version_code)
    *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, 0);

  fp = fopen (PATH_METADL_CONFIG, "r");
  if (!fp)
    return SANE_STATUS_GOOD;	/* don't insist on config file */

  while (fgets (backend_name, sizeof (backend_name), fp))
    {
      if (backend_name[0] == '#')	/* ignore line comments */
	continue;
      len = strlen (backend_name);
      if (backend_name[len - 1] == '\n')
	backend_name[--len] = '\0';

      if (!len)
	continue;		/* ignore empty lines */

      add_backend (backend_name, 0);
    }
  fclose (fp);

  return SANE_STATUS_GOOD;
}

void
sane_exit (void)
{
  struct backend *be, *next;

  for (be = first_backend; be->next; be = next)
    {
      next = be->next;
      if (be->loaded)
	{
	  (*be->op[OP_EXIT]) ();
	  dlclose (be->handle);
	}
      if (be->name)
	free ((void *) be->name);
      free (be);
    }
  first_backend = 0;
}

/* Note that a call to get_devices() implies that we'll have to load
   all backends.  To avoid this, you can call sane_open() directly
   (assuming you know the name of the backend/device).  This is
   appropriate for the command-line interface of SANE, for example.
 */
SANE_Status
sane_get_devices (const SANE_Device *** device_list)
{
  static int devlist_size = 0, devlist_len = 0;
  static const SANE_Device **devlist;
  const SANE_Device **be_list;
  struct backend *be;
  SANE_Status status;
  char *full_name;
  int i, num_devs;
  size_t len;
#define ASSERT_SPACE(n)						\
  {									\
    if (devlist_len + (n) > devlist_size)				\
      {									\
	devlist_size += (n) + 15;					\
	devlist = realloc(devlist, devlist_size * sizeof(devlist[0]));	\
	if (!devlist)							\
	  return SANE_STATUS_NO_MEM;					\
      }									\
  }

  if (devlist)
    for (i = 0; i < devlist_len; ++i)
      free ((void *) devlist[i]);
  devlist_len = 0;

  for (be = first_backend; be; be = be->next)
    {
      if (!be->loaded)
	load (be);

      status = (long) (*be->op[OP_GET_DEVS]) (&be_list);
      if (status != SANE_STATUS_GOOD)
	continue;

      /* count the number of devices for this backend: */
      for (num_devs = 0; be_list[num_devs]; ++num_devs);

      ASSERT_SPACE (num_devs);

      for (i = 0; i < num_devs; ++i)
	{
	  SANE_Device *dev;
	  char *mem;

	  /* create a new device entry with a device name that is the
	     sum of the backend name a colon and the backend's device
	     name: */
	  len = strlen (be->name) + 1 + strlen (be_list[i]->name);
	  mem = malloc (sizeof (*dev) + len + 1);
	  if (!mem)
	    return SANE_STATUS_NO_MEM;

	  full_name = mem + sizeof (*dev);
	  strcpy (full_name, be->name);
	  strcat (full_name, ":");
	  strcat (full_name, be_list[i]->name);

	  dev = (SANE_Device *) mem;
	  dev->name = full_name;
	  dev->vendor = be_list[i]->vendor;
	  dev->model = be_list[i]->model;
	  dev->type = be_list[i]->type;

	  devlist[devlist_len++] = dev;
	}
    }

  /* terminate device list with NULL entry: */
  ASSERT_SPACE (1);
  devlist[devlist_len++] = 0;

  *device_list = devlist;
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_open (SANE_String_Const full_name, SANE_Handle * meta_handle)
{
  const char *be_name, *dev_name;
  struct meta_scanner *s;
  SANE_Handle *handle;
  struct backend *be;
  SANE_Status status;

  dev_name = strchr (full_name, ':');
  if (dev_name)
    {
#ifdef strndupa
      be_name = strndupa (full_name, dev_name - full_name);
#else
      char *tmp;

      tmp = alloca (dev_name - full_name + 1);
      memcpy (tmp, full_name, dev_name - full_name);
      tmp[dev_name - full_name] = '\0';
      be_name = tmp;
#endif
      ++dev_name;		/* skip colon */
    }
  else
    {
      /* if no colon interpret full_name as the backend name; an empty
         backend device name will cause us to open the first device of
         that backend.  */
      be_name = full_name;
      dev_name = "";
    }

  for (be = first_backend; be; be = be->next)
    if (strcmp (be->name, be_name) == 0)
      break;

  if (!be)
    {
      status = add_backend (be_name, &be);
      if (status != SANE_STATUS_GOOD)
	return status;
    }

  if (!be->loaded)
    {
      status = load (be);
      if (status != SANE_STATUS_GOOD)
	return status;
    }

  status = (long) (*be->op[OP_OPEN]) (dev_name, &handle);
  if (status != SANE_STATUS_GOOD)
    return status;

  s = malloc (sizeof (*s));
  if (!s)
    return SANE_STATUS_NO_MEM;

  memset (s, 0, sizeof (*s));
  s->be = be;
  s->handle = handle;
  *meta_handle = s;

  return SANE_STATUS_GOOD;
}

void
sane_close (SANE_Handle handle)
{
  struct meta_scanner *s = handle;

  (*s->be->op[OP_CLOSE]) (s->handle);
  free (s);
}

const SANE_Option_Descriptor *
sane_get_option_descriptor (SANE_Handle handle, SANE_Int option)
{
  struct meta_scanner *s = handle;
  return (*s->be->op[OP_GET_OPTION_DESC]) (s->handle, option);
}

SANE_Status
sane_control_option (SANE_Handle handle, SANE_Int option,
		     SANE_Action action, void *value, SANE_Word * info)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_CTL_OPTION]) (s->handle, option, action,
					     value, info);
}

SANE_Status
sane_get_parameters (SANE_Handle handle, SANE_Parameters * params)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_GET_PARAMS]) (s->handle, params);
}

SANE_Status
sane_start (SANE_Handle handle)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_START]) (s->handle);
}

SANE_Status
sane_read (SANE_Handle handle, SANE_Byte * data, SANE_Int max_length,
	   SANE_Int * length)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_READ]) (s->handle, data, max_length, length);
}

void
sane_cancel (SANE_Handle handle)
{
  struct meta_scanner *s = handle;
  (*s->be->op[OP_CANCEL]) (s->handle);
}

SANE_Status
sane_set_io_mode (SANE_Handle handle, SANE_Bool non_blocking)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_SET_IO_MODE]) (s->handle, non_blocking);
}

SANE_Status
sane_get_select_fd (SANE_Handle handle, SANE_Int * fd)
{
  struct meta_scanner *s = handle;
  return (long) (*s->be->op[OP_GET_SELECT_FD]) (s->handle, fd);
}
